"""
Early draft of a very handy utility that takes a CCDL file (Concise Campaign Definition Language)
and creates a graph(viz) visualization of it.
"""
import sys
import graphviz # This makes emod-api dependent on graphviz. Might be nice if this was optional?

debug = False
def get_nickname_from_event( event_num, pieces ):
    event_name = f"[{event_num}]{pieces[0]}*{pieces[1]}*{pieces[2]}"
    """
    Allow nodes to get briefer and potentially more helpful nicknames. Default will probably remain a
    the nasty autogen above. Users can override this function with a callback of their own.
    """

    # If user has manually added an optional 5th 'column', use that, but turns certain characters into newlines.
    if len(pieces)==5:
        event_name = pieces[4].strip().replace( " ", "\n" ).replace("->","%%%").replace( "-", "\n" ).replace( "%%%", "->" )
    return event_name 


def get_colour_from_event( tokens ): 
    """
    Allow nodes to get a content-dependent colour. Default to just white. Users can override this function with 
    a callback of their own. Have been using colour to capture IP categories. 
    """
    return "white" 

def get_shape_from_event( tokens ):
    """
    Allow nodes to get a content-dependent shape. Default to circle. Users can override this function with 
    a callback of their own. Have been using shape to capture 'epoch' categories. 
    """
    return "circle" 


def set_beautifiers( name_cb=None, colour_cb=None, shape_cb=None ):
    """
    Override default no-op callbacks for setting nicknames, colours, and shapes of campaign nodes
    """
    global get_nickname_from_event, get_colour_from_event, get_shape_from_event
    if name_cb:
        get_nickname_from_event = name_cb
    if colour_cb:
        get_colour_from_event = colour_cb
    if shape_cb:
        get_shape_from_event = shape_cb

def viz( in_name="campaign.ccdl", out_name = "camp.sv", display = True ): 
    ccdl = open( in_name ).readlines()
    dot = graphviz.Digraph(out_name, comment='Patient Pathway Design')

    producers = {}
    consumers = {}
    broadcasters = ["BroadcastEvent", "SimpleHealthSeekingBehavior", "DiagnosticTreatNeg" ]

    event_names = {}
    event_num = 0
    for camp_event in ccdl:
        pieces = camp_event.strip().split( "::" )
        if len(pieces)<2:
            event_num += 1
            continue

        if "Outbreak" in pieces[-1]:
            # It is useful to 'hack' a few lines so that major infection lifecycle events get tied into
            # our visualization.
            camp_event = camp_event.replace( pieces[-1], pieces[-1]+"+BroadcastEvent(TBActivation)" )
            pieces = camp_event.strip().split( "::" )

        if debug:        
            print( f"Processing line {camp_event}." )

        # Color represents IP
        # Shape represents Epoch
        node_name = get_nickname_from_event( event_num, pieces )
        event_names[event_num] = node_name
        node_color = get_colour_from_event(pieces)
        node_shape = get_shape_from_event(pieces)

        dot.node( name=event_names[event_num], style='filled', fillcolor = node_color, shape=node_shape )
        if debug:
            print( f"Creating node with name {event_names[event_num]}." )
        triggered = pieces[3].split( "->" )
        if len(triggered)>1:
            triggers = triggered[0].strip()
            for trigger in triggers.split( "+" ):
                if trigger not in consumers:
                    consumers[trigger] = set() 
                consumers[trigger].add( event_num )

        for broadcaster in broadcasters:
            sender = (triggered[-1].split( "+" )[-1]).strip()
            if broadcaster in sender:
                from parse import parse
                #regex = "{}" + broadcaster + "({}){}"
                regex = broadcaster + "({})"
                #print( f"Looking for {regex} in {sender}" )
                #tokens = parse( regex, triggered[1] )
                tokens = parse( regex, sender )
                signals = tokens[0]
                for signal in signals.split( "/" ):
                    #print( f"Found an event broadcasting {signal}" )
                    if signal not in producers:
                        producers[signal] = set()
                    producers[signal].add( event_num )
        event_num += 1

    if debug:
        print( "PRODUCERS\n" )
        print( producers )
        print( "\nCONSUMERS\n" )
        print( consumers )
    event_whitelist = []
    if len(sys.argv)>2:
        event_whitelist.append( sys.argv[2] )

    for event in producers:
        if event_whitelist and event not in event_whitelist:
            continue
        senders = producers[event]
        if event in consumers:
            receivers = consumers[event]
            for from_event in senders:
                for to_event in receivers:
                    ip_from = ccdl[from_event].split( "::" )[2].split("/")[1] if "/" in ccdl[from_event].split( "::" )[2] else "*"
                    ip_to = ccdl[to_event].split( "::" )[2].split("/")[1] if "/" in ccdl[to_event].split( "::" )[2] else "*"
                    #if ip_to != "*" and ip_from != "*" and ip_from.split(":")[0] == ip_to.split(":")[0] and ip_from.split(":")[1] != ip_to.split(":")[1]:
                    if ip_to != "*" and ip_from != "*":
                        ip_from_key = ip_from.split("=")[0]
                        ip_from_value = ip_from.split("=")[1]
                        ip_to_key = ip_to.split("=")[0]
                        ip_to_value = ip_to.split("=")[1]
                        if ip_from_key == ip_to_key and ip_from_value != ip_to_value:
                            continue # aka do not continue
                    time_from = ccdl[from_event].split( "::" )[0]
                    if "(" in time_from:
                        time_from = time_from.split( "(" )[0]
                    time_to = ccdl[to_event].split( "::" )[0]
                    time_from_start = time_from
                    time_to_start = time_to
                    time_from_end = 1e9
                    time_to_end = 1e9
                    if "-" in time_from:
                        time_from_start = float(time_from.split("-")[0])
                        time_from_end = float(time_from.split("-")[1])
                    if "-" in time_to:
                        time_to_start = float(time_to.split("-")[0])
                        time_to_end = float(time_to.split("-")[1])
                    time_from_start = float(time_from_start)
                    time_to_start = float(time_to_start)
                    if time_from_end <= time_to_start or time_from_start >= time_to_end:
                        if debug:
                            print( f"Discarding edge from {from_event} to {to_event} because not overlapping in time." )
                            print( time_from_start, time_from_end, time_to_start, time_to_end )
                        continue
                    if debug:
                        #print( f"Creating edge from {from_event} to {to_event} for {event}." )
                        #print( f"{from_event} ---{event} ---> {to_event}" )
                        print( f"{event_names[from_event]} ---{event} ---> {event_names[to_event]}" )
                    #dot.edge( from_event, to_event, name=event )
                    dot.edge( str(event_names[from_event]), str(event_names[to_event]), label=event )
    dot

    #dot.render()
    #if display:
    #    from graphviz import Source
    #    path = out_name + ".gv"
    #    s = Source.from_file(path)
    #    s.view()

if __name__ == "__main__":
    viz()
